%{
open Fc_ast_type
open Fc_ast_util

(*
 * Build a binary expression.
 *)
let make_binop op expr1 expr2 =
   let pos = union_pos (pos_of_expr expr1) (pos_of_expr expr2) in
      BinOpExpr (op, expr1, expr2, pos)

(*
 * TODO: put your utilities here.
 *)
%}

/*
 * Example binary opertors.
 * We have the binary ops return
 * their position in the file.
 */
%token <Fc_ast_type.pos> TokPlus
%token <Fc_ast_type.pos> TokMinus
%token <Fc_ast_type.pos> TokStar
%token <Fc_ast_type.pos> TokSlash

/*
 * Example keywords.  Same deal, the
 * keywords return their position.
 */
%token <Fc_ast_type.pos> TokTypedef
%token <Fc_ast_type.pos> TokIf
%token <Fc_ast_type.pos> TokElse
%token <Fc_ast_type.pos> TokFor
%token <Fc_ast_type.pos> TokWhile
%token <Fc_ast_type.pos> TokReturn
%token <Fc_ast_type.pos> TokBreak
%token <Fc_ast_type.pos> TokStruct

/*
 * Some symbols.
 */
%token <Fc_ast_type.pos> TokSemi
%token <Fc_ast_type.pos> TokLeftParen
%token <Fc_ast_type.pos> TokRightParen
%token <Fc_ast_type.pos> TokLeftBrack
%token <Fc_ast_type.pos> TokRightBrack
%token <Fc_ast_type.pos> TokLeftBrace
%token <Fc_ast_type.pos> TokRightBrace

/*
 * Other stuff.
 */
%token <Fc_ast_type.pos> TokEof

/*
 * Names, numbers, etc.
 */
%token <Symbol.symbol * Fc_ast_type.pos> TokId
%token <Symbol.symbol * Fc_ast_type.pos> TokTypeId
%token <string * Fc_ast_type.pos> TokString
%token <char * Fc_ast_type.pos> TokChar
%token <int * Fc_ast_type.pos> TokInt
%token <float * Fc_ast_type.pos> TokFloat

/*
 * Declare the "start" productions.
 * These declarations define functions for parsing
 * the productions.  You can define as many as you want.
 */
%start prog
%start topstmt
%type <Fc_ast_type.prog> prog
%type <Fc_ast_type.expr option> topstmt

%%

/*
 * TODO: implement the syntax for a prog.
 * This defines the valid contents of a file.
 */
prog:	  TokEof
          { raise (Invalid_argument "Fc_ast_parse.prog: not implemented") }
        ;

/*
 * Toplevel statement.  For evaluation,
 * we allow type, function, and value definitions
 * at the top level.
 */
topstmt:  stmt
          { Some $1 }
        | TokEof
          { None }
        ;

/*
 * Some example expressions.
 */
expr:     TokInt
          { let i, pos = $1 in
               IntExpr (i, pos)
          }
        | TokChar
          { let c, pos = $1 in
               CharExpr (c, pos)
          }
        | TokFloat
          { let x, pos = $1 in
               FloatExpr (x, pos)
          }
        | TokString
          { let s, pos = $1 in
               StringExpr (s, pos)
          }
        | expr TokPlus expr
          { make_binop PlusOp $1 $3 }
        | expr TokMinus expr
          { make_binop MinusOp $1 $3 }
        
          /* You have to implement a lot of other stuff here */
        ;

/*
 * Some example statements.
 */
stmt:     expr TokSemi
          { $1 }
        | TokIf TokLeftParen expr TokRightParen stmt
          { let pos = union_pos $1 (pos_of_expr $5) in
               IfExpr ($3, $5, None, pos)
          }
        | TokIf TokLeftParen expr TokRightParen stmt TokElse stmt
          { let pos = union_pos $1 (pos_of_expr $7) in
               IfExpr ($3, $5, Some $7, pos)
          }

          /* Lots of other stuff here */
        ;
